from matplotlib.offsetbox import TextArea, VPacker, AnnotationBbox
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.patches as mpatches
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
from datetime import datetime
from matplotlib.pylab import *
import matplotlib.cm as cm
from matplotlib import rc
import pandas as pd
import numpy as np
from math import *
import colorsys
import csv


def read_file(start_date, end_date, file):
    """ Read and transform a csv file into a dataframe
            Parameters:
                - start_date(str): the reading start date
                - end_date(str): the reading end date
                - file(csv): the reading file
            Returns:
                - df(dataframe): the dataframe that contains the csv file
                - Month(list):  thereading month
    """
    df = pd.read_csv(file, encoding='latin', delimiter=",",
                     decimal=',', index_col=0, parse_dates=True)
    df = df[start_date:end_date]
    df = df.apply(pd.to_numeric, errors='coerce')
    Month = df.index.month
    return df, Month


def clothing_insulation(Month):
    """ Calculates the clothing insulation for a given month
            Parameter:
                - Month(list): the studied month
            Return:
                - Icl(list): clothing insulation
    """
    Icl = []
    summer = [5, 6, 7, 8, 9]
    winter = [1, 2, 3, 4, 10, 11, 12]
    for i in Month:
        if i in winter:
            Icl.append(1*0.155)
        else:
            Icl.append(0.5*0.155)
    return np.array(Icl)


def clothing_surface_factor(Icl):
    """ Calculates the clothing surface area factor according to
            the values of clothing insulation
            Parameter:
                - Icl(list): clothing insulation
            Return:
                - fcl(list): clothing surface area factor
    """
    fcl = np.ones(len(Icl))
    fcl[Icl > 0.078] = 1.05+0.645*Icl[Icl > 0.078]
    fcl[Icl <= 0.078] = 1.00+1.290*Icl[Icl <= 0.078]
    return fcl


def relative_air_velocity(M, va):
    """Calculates the relative air velocity
            Parameters:
                - M(float): metabolic rate
                - va(float): air velocity
            Return:
                - var(float): relative air velocity
    """
    var = va+0.0052*(M-58)
    return var


def convective_heat(var, va):
    """Calculates the convective heat transfer coefficient
            Parameters:
                - var(float): relative air velocity
                - va(float): air velocity
            Return :
                - hc(float): convective heat transfer coefficient
    """
    if va <= 1:
        hc = 3.5+5.2*var
    else:
        hc = 8.7*var**(0.6)
    return hc


def clothing_surface_temperature(M, Icl, ta, fcl, hc, tr, W):
    """Calculates the  clothing surface temperature
            Parameters:
                - M(float): metabolic rate
                - Icl(list): clothing insulation
                - ta(list):  air temperature
                - fcl(list): clothing surface factor
                - hc(float): convective heat transfer coefficient
                - tr(list): mean radiant temperature
                - W(float): effective mechanical power
            Return :
                - tcl(list): clothing surface temperature
    """
    MW = M-W
    list_tcl = []
    for j in range(len(ta)):
        Y = -Icl[j]*3.96*(10**(-8)*fcl[j])
        a0 = -Y
        a1 = -Y*4*273
        a2 = -Y*6*273**2
        a3 = (1+Icl[j]*fcl[j]*hc-Y*4*273**3)
        x = Icl[j]*fcl[j]*hc*ta[j]
        a4 = 35.7-0.028*MW+Icl[j]*(3.96*(10**(-8)*fcl[j]*(tr[j]+273)**4))+Y*273**4+x
        coeff = [a0, a1, a2, a3, -a4]
        co = np.roots(coeff)
        for i in co:
            if i > 0 and i < 70:
                tcl = i
        list_tcl.append(tcl.real)
    return np.array(list_tcl, dtype=float)


def water_vapor_pressure(ta, hum):
    """ Calculates the  water vapor partial pressure
            Parameters:
                - ta(list): air temperature
                - hum(list): humidity
            Return :
                - pa(list): water vapor partial pressure
    """
    psat = np.ones(ta.shape)
    psat[ta < 0] = 611*np.exp((22.44*ta[ta < 0])/(272.44+ta[ta < 0]))
    psat[ta >= 0] = 611*np.exp((17.08*ta[ta >= 0])/(234.18+ta[ta >= 0]))
    pa = (psat*hum)/100
    return pa


def PMV(M, pa, tcl, ta, fcl, hc, tr, W):
    """ Calculates the  PMV
            Parameters:
                - M(float): metabolic rate
                - pa(list): water vapor partial pressure
                - tcl(list): clothing surface temperature
                - ta(list): air temperature
                - fcl(list): clothing surface factor
                - hc(float): convective heat transfer coefficient
                - tr(list): mean radiant temperature
                - W(float): effective mechanical power
            Return :
                - pmv(list): PMV
    """
    MW = M-W
    x1 = 0.303*exp(-0.036*M)+0.028
    x2 = 5733-6.995*(MW)-pa
    x3 = -1.7*10**(-5)*M*(5867-pa)-0.0014*M*(34-ta)
    x4 = (tcl+273)**4-(tr+273)**4
    x5 = fcl*hc*(tcl-ta)
    pmv = x1*(MW-3.05*10**(-3)*x2-0.42*(MW-58.15)+x3-3.96*10**(-8)*fcl*x4-x5)
    return pmv


def text(pos_x, pos_y, ax, fig, num_elem, month, M, va, elem):
    """ Adds text to the figure
            Parameters:
            - pos_x(float): position of the text according to the x-axis
            - pos_y(float): position of the text according to the y-axis
            - ax, fig: the figure that we add text
            - num_elem(int): number of the node or zone studied
            - month (int): number of month studied
            - va (float): air velocity
            - elem(str): zone or node studied
    """
    texts = [elem + "=" + str(num_elem), "month={}".format(month),
             'M = {}'.format(M), "va = {}".format(va), "Tr = ta", "W = 0"]
    Texts = []
    for t in texts:
        Texts.append(TextArea(t))
    texts_vbox = VPacker(children=Texts, pad=0, sep=0)
    ann = AnnotationBbox(texts_vbox, (pos_x, pos_y), xycoords=ax.transAxes,
                         box_alignment=(0.3, .1),
                         bboxprops=dict(facecolor='white', boxstyle='round'))
    ann.set_figure(fig)
    fig.artists.append(ann)


def PPD(PMV):
    """ Calculates the ppd indice
            Parameter :
                - PMV(list): pmv indice
            Return :
                - ppd(list): ppd indice
    """
    ppd = 100-95*np.exp(-0.03353*PMV**4-0.2179*PMV**2)
    return ppd


def plot_contour(pmv, ppd, ta, hum, M, num_elem, month, va, elem):
    """ plot the contour of pmv  as a function of temperature and humidity.
            Parameters:
                - pmv(list): pmv indice
                - ppd(list): ppd indice
                - ta(list):  air temperature
                - hum(list): humidity
                - M(float): metabolic rate
                - num_elem(int): number of the node or zone studied
                - month (int): number of month studied
                - va (float): air velocity
                - elem(str): zone or node studied
    """
    fig, ax = plt.subplots(figsize=(6, 6))
    plt.subplots_adjust(left=0.12, bottom=0.11, right=0.6, top=0.94)
    contours = plt.contour(ta, hum, pmv, 3, colors='black')
    plt.clabel(contours, inline=True, fontsize=12)
    plt.imshow(pmv, extent=[np.min(ta)-0.5, np.max(ta)+0.5,
                            np.min(hum), np.max(hum)],
               origin='lower', cmap='RdGy', alpha=0.5)
    plt.colorbar()
    h1, _ = contours.legend_elements()
    plt.legend([h1[0]], ['PMV'], loc='best')
    ax.set_title('PMV')
    ax.set_xlabel('Temperature(°C)')
    ax.set_ylabel('Humidity(%)')
    text(1.8, 0.48, ax, fig, num_elem, month, M, va, elem)
    plt.savefig("comfort_png/contour {}.png".format(month))


def comfort_thermique(M, va, month, ta, hum, tr, W, ta_tcl):
    """Calculates the thermal comfort
            Parameters:
                - M(float): metabolic rate
                - ta(list): air temperature
                - hum(list): humidity
                - month (int): number of studied month
                - ta_tcl(list): air temperature
                - tr(list): mean radiant temperature
                - W(float): effective mechanical power
            Return :
                - pmv(list): pmv indice
                - ppd(list): ppd indice
    """
    var = relative_air_velocity(M, va)
    hc = convective_heat(var, va)
    Icl = clothing_insulation(month)
    fcl = clothing_surface_factor(Icl)
    tcl = clothing_surface_temperature(M, Icl, ta_tcl, fcl, hc, tr, W)
    pa = water_vapor_pressure(ta, hum)
    pmv = PMV(M, pa, tcl, ta, fcl, hc, tr, W)
    ppd = PPD(pmv)
    return pmv, ppd


def comfort_level(pmv):
    """Calculates the comfort level of pmv
            Parameter:
                - pmv(list): pmv indice
            Return:
                - comf(list): comfort level
    """
    comf = []
    for i in range(len(pmv)):
        if(pmv[i] >= -3 and pmv[i] < -2):
            comf.append(0)
        if (pmv[i] >= -2 and pmv[i] < -1):
            comf.append(1)
        if (pmv[i] >= -1 and pmv[i] < -0.5):
            comf.append(2)
        if (pmv[i] >= -0.5 and pmv[i] <= 0.5):
            comf.append(3)
        if (pmv[i] > 0.5 and pmv[i] < 1):
            comf.append(4)
        if (pmv[i] >= 1 and pmv[i] <= 2):
            comf.append(5)
        if (pmv[i] >= 2 and pmv[i] <= 3):
            comf.append(6)
    return comf


def list_colors(comf, pmv):
    """ Assigns for each degree of comfort its colour and comfort scale
            Parameters:
                - pmv(list): pmv indice
                - comf(list): comfort level
            Retourns:
                - colors(list): color according to comfort level
                - scale_bedford(list): the color scale
    """
    comf = list(set(comf))
    comf = np.sort(comf)
    colors = []
    dic_colors = {"#0103fb": {"scale": [-3, -2],
                              "bedford": "Too cold"},
                  "#014e99": {"scale": [-2, -1],
                              "bedford": "cold"},
                  "#5ba9fa": {"scale": [-1, -0.5],
                              "bedford": "Slightly cold"},
                  "#0af505": {"scale": [-0.5, 0.5],
                              "bedford": " Comfortably "},
                  "#ff0a37": {"scale": [0.5, 1],
                              "bedford": " Slightly warm "},
                  "#bb2505": {"scale": [1, 2],
                              "bedford": "Warm"},
                  "#fc0405": {"scale": [2, 3],
                              "bedford": "Too hot"}, }
    scale = []
    bedford = []
    key = list(dic_colors.keys())
    for i in range(len(comf)):
        colors.append(key[comf[i]])
        scale.extend(dic_colors[colors[-1]]["scale"])
        bedford.append(dic_colors[colors[-1]]["bedford"])
    scale = list(set(scale))
    scale = np.sort(scale)
    scale = np.array(scale, dtype=float64)
    scale[0] = float(np.round(np.min(pmv), 2))
    scale[-1] = np.round(np.max(pmv), 2)
    scale_bedford = []
    for s, b in zip(scale, bedford):
        scale_bedford.extend([s, b])
    scale_bedford.append(scale[-1])
    return colors, scale_bedford


def plot_ppd(pmv, ppd, M,  num_elem, month, va, elem):
    """ plot the ppd  as a function of pmv.
            Parameters:
                - pmv(list): pmv indice
                - ppd(list): ppd indice
                - M(float): metabolic rate
                - num_elem(int): number of node or zone studied
                - month (int): number of month studied
                - va (float): air velocity
                - elem(str): zone or node studied
    """
    fig, ax = plt.subplots(figsize=(10, 10))
    comf = comfort_level(pmv)
    colors, scale = list_colors(comf, pmv)
    cmaps = matplotlib.colors.ListedColormap(colors)
    scatter = ax.scatter(x=pmv, y=ppd, c=comf, cmap=cmaps)
    if len(scale) == 3:
        cb = plt.colorbar(scatter, ticks=[comf[0]])
        cb.ax.set_yticklabels(["{} \n [{}, {}]".format(scale[1],
                               scale[0], scale[2])])
    else:
        pas = (np.max(comf) - np.min(comf))/(2*len(colors))
        tick = np.arange(np.min(comf), np.max(comf)+pas, pas)
        cb = plt.colorbar(scatter, ticks=tick)
        cb.ax.set_yticklabels(scale)
    cb.set_label('PMV')
    y = np.full(len(pmv), 10)
    plt.plot(pmv, y, label="PPD=10")
    plt.legend()
    ax.set_title('PPD as a function of PMV')
    ax.set_xlabel('PMV')
    ax.set_ylabel('PPD(%)')
    text(0.9, 0.8, ax, fig,  num_elem, month, M, va, elem)
    plt.grid()
    plt.savefig("comfort_png/ppd-PMV-month{}.png".format(month))


def date(df):
    """ Changes the date format of the dataframe
            Parameter:
                - df(dataframe): contains the studied data
            Returns:
                -  df(dataframe): contains the studied data where
                                  the date format is changed
                -  data(list): contains the date that will appear on the plot
    """
    df["date"] = pd.to_datetime(df.index)
    df["date"] = df["date"].dt.strftime("%m-%d-%H"+"h")
    df = df.set_index('date')
    x = np.linspace(0, df.shape[0]-1, 10)
    x = np.array(x, dtype=int)
    date = []
    for i in range(df.shape[0]):
        if i in x:
            date.append(df.index[i])
    return df, date


def plot_time(pmv, ppd, ta, hum, df, num_elem, month, M, va, elem):
    """ Plots the fields in relation to a tomporal series
            Parameters:
                - pmv(list): pmv indice
                - ppd(list): ppd indice
                - ta(list):  air temperature
                - hum(list): humidity
                - df(dataframe): contains the studied data
                - M(float): metabolic rate
                - num_elem(int): number of node or zone studied
                - month (int): number of month studied
                - va (float): air velocity
                - elem(str): zone or node studied
    """
    df, time = date(df)
    unit = ["PMV", "PPD(%)", "Temperature(°C)", "Humidity(%)"]
    comf = comfort_level(pmv)
    colors, scale = list_colors(comf, pmv)
    cmaps = matplotlib.colors.ListedColormap(colors)
    list_plot = [pmv, ppd, ta, hum]
    titre = ["PMV", "ppd", "temperature", "humidity"]
    i = 0
    for plot in list_plot:
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_subplot(1, 1, 1)
        plt.subplots_adjust(right=0.79)
        scatter = plt.scatter(x=df.index, y=plot, c=comf, cmap=cmaps)
        if len(scale) == 3:
            cb = plt.colorbar(scatter, ticks=[comf[0]])
            cb.ax.set_yticklabels(["{} \n [{}, {}]".format(scale[1],
                                   scale[0], scale[2])])
        else:
            pas = (np.max(comf) - np.min(comf))/(2*len(colors))
            tick = np.arange(np.min(comf), np.max(comf)+pas, pas)
            cb = plt.colorbar(scatter, ticks=tick)
            cb.ax.set_yticklabels(scale)
        text(0.9, 0.85, ax, fig,  num_elem, month, M, va, elem)
        cb.set_label('PMV')
        plt.xticks(rotation='vertical')
        ax.set_xticks(time)
        plt.ylabel(titre[i])
        plt.grid()
        ax.set_title('representation of {} as a function of time'
                     .format(titre[i]))
        ax.set_xlabel('Date')
        ax.set_ylabel('{}'.format(unit[i]))
        fig.autofmt_xdate(rotation=45)
        plt.savefig("comfort_png/{}-month {}.png"
                    .format(titre[i], month))
        i = i+1


def nodes_to_zones(file, zone_to_request):
    """ Reads json file.
            Parameters:
                    file (string) : the json file to read
                    zone_to_request(list of string) : list of zones
            Returns:
                    my_dic(dict): contains the nodes of each zone and
                                  their area
    """
    df = pd.read_json(file)
    my_dic = {}
    for zone in zone_to_request:
        if(int(df.building[zone]['n_nodes'])) != 0:
            zone_node_area = []
            for i in range(1, int(df.building[zone]['n_nodes'])+1):
                zone_node_area.append(("zigduino-"+str(df.building[zone]
                                       ['node'+str(i)]['node_id']),
                                       float(df.building[zone]['node'+str(i)]
                                       ['location']['space_area'])))
            my_dic.update({str(zone): zone_node_area})
        else:
            print("No nodes in", zone)
    return my_dic


def aggregatedCsv(file, var_temp, aggregated_dict):
    ''' Creates a csv file that contains var_temp and the aggregated_dict
            Parameters:
                file (string) : the csv file to create
                var_temp(list of string) : list of date_format
                aggregated_dict(dict):  contains data aggregated by zones
    '''
    with open(file, "w", newline='') as my_csv:
        writer = csv.writer(my_csv)
        first_ligne = ["date_format"]
        aggregated_data = []
        for key, value in aggregated_dict.items():
            for ke, val in value.items():
                first_ligne += [key+"-"+ke]
                aggregated_data.append(val)
        var_temp.extend(aggregated_data)
        writer.writerow(first_ligne)
        writer.writerows(zip(*var_temp))


def mean_aggregated(aggregated_dict, zone, field, list_value, value_field):
    '''Calculates the average weighted by the surfaces of the different
                                 rooms within a given area
            Parameters:
                    zone(str): affected zone
                    field(str): affected field
                    list_value(tuple): tuple of node with its area
                    value_field(list of list): list of values of the field
                                                of the different nodes
                                                associated to the zone
            Returns:
                    aggregated_dict(dict): contains data aggregated by zones
    '''
    area = [v for n, v in list_value]
    mean = []
    if not(aggregated_dict.get(zone)):
        aggregated_dict[zone] = {}
    for j in range(len(value_field[0])):
        sum_area = 0
        sum_values = 0
        for i in range(len(area)):
            sum_area += area[i]
            try:
                sum_values += (area[i] * value_field[i][j])
            except Exception:  # in case different length of value_field
                continue
        mean.append(sum_values / sum_area)

    aggregated_dict[zone][field] = mean
    return aggregated_dict


def aggregated(csv_file, zone_to_request, df, fields):
    """ Aggregates the csv_file elements according to the zone(s)
            to request and the field(s)
            Parameters:
                - csv_file(csv): contains the elements to aggregate
                - zone_to_request(list): the zone(s) to aggregate
                - field(list): field to aggregate
                - df(dataframe): contains the elemets to aggregate
            Returns:
                - df_aggregated(dataframe): contains aggregated elemets
    """
    dic_zones = nodes_to_zones("multizone_modele_attributes.json",
                               zone_to_request)
    aggregated_dict = {}
    temp = [df.index.values]
    for field in set(fields):
        for key, value in dic_zones.items():
            val_field = []
            for node, area in value:
                val = [temp for temp in csv_file[node+":"+field]]
                val_field.append(val)
            aggregated_dict = mean_aggregated(aggregated_dict, key,
                                              field, value, val_field)
    aggregatedCsv("../results_csv/aggregated_comfort.csv", temp,
                  aggregated_dict)
    df_aggregated, month = read_file(None, None,
                                     "../results_csv/aggregated_comfort.csv")
    return df_aggregated


if '__main__':
    nb_ech = 100
    M = 1.2*58
    W = 0
    va = 0.1
    start_date = "2019-07-01T00:00:00"
    end_date = "2019-07-31T23:00:00"
    file_inpute = '../results_csv/export_comfort.csv'
    node = 40
    zone = [13]
    field_to_request = ["temperature", "humidity", "pmv", "ppd"]
    nodes = [19, 53, 54, 40, 11, 17, 39, 26, 45, 52, 55, 56, 23,
             24, 44, 46, 48, 48, 50, 51, 11, 16,  22, 47]
    df, month = read_file(start_date, end_date, file_inpute)
    choices = "zone"
    if choices == "zone":
        for node in nodes:
            ta = df.loc[:, "zigduino-{}:temperature".format(node)].values
            hum = df.loc[:, "zigduino-{}:humidity".format(node)].values
            pmv, ppd = comfort_thermique(M, va, month, ta, hum, ta, W, ta)
            df["zigduino-{}:pmv".format(node)] = pmv
            df["zigduino-{}:ppd".format(node)] = ppd
        df.to_csv("../results_csv/comfort_pmv.csv")
        csv_file = pd.read_csv("../results_csv/comfort_pmv.csv")
        zone_to_request = ["zone{}".format(i) for i in zone]
        df_aggregated = aggregated(csv_file, zone_to_request, df,
                                   field_to_request)
        ta = df_aggregated.loc[:, "zone{}-temperature".format(zone[0])].values
        hum = df_aggregated.loc[:, "zone{}-humidity".format(zone[0])].values
        pmv = df_aggregated.loc[:, "zone{}-pmv".format(zone[0])].values
        ppd = df_aggregated.loc[:, "zone{}-ppd".format(zone[0])].values
        num_elem = zone[0]
    else:
        ta = df.loc[:, "zigduino-{}:temperature".format(node)].values
        hum = df.loc[:, "zigduino-{}:humidity".format(node)].values
        pmv, ppd = comfort_thermique(M, va, month, ta, hum, ta, W, ta)
        num_elem = node
    list_ta = np.linspace(np.min(ta), np.max(ta), nb_ech)
    list_hum = np.linspace(np.min(hum), np.max(hum), nb_ech)
    ta_mesh, hum_mesh = np.meshgrid(list_ta, list_hum)
    list_pmv, list_ppd = comfort_thermique(M, va, month[:nb_ech], ta_mesh,
                                           hum_mesh, list_ta, W, ta_mesh[0, :])
    plot_contour(list_pmv, list_ppd, ta_mesh, hum_mesh, M, num_elem,
                 month[0], va, choices)
    list_pmv, list_ppd = comfort_thermique(M, va, month[:nb_ech], list_ta,
                                           list_hum, list_ta, W, list_ta)
    plot_ppd(list_pmv, list_ppd, M, num_elem, month[0], va, choices)
    comf = comfort_level(pmv)
    plot_time(pmv, ppd, ta, hum, df, num_elem, month[0], M, va, choices)
